import inspect

from munch import Munch


class PropertyMunch(Munch):
    def __init__(self, instance, *args, **kwargs):
        """
        A Munch dictionary that dynamically exposes only @property attributes
        while keeping normal dictionary behavior for other keys.

        :param instance: The object whose properties will be dynamically accessible.
        """
        super().__init__(*args, **kwargs)
        self._instance = instance  # Store the actual object

        # Inject only properties (methods decorated with @property)
        for attr, value in inspect.getmembers(type(instance), lambda x: isinstance(x, property)):
            self._inject_property(attr)

    def _inject_property(self, attr):
        """
        Injects a @property as a dynamic property into the Munch dictionary.
        """
        prop = property(
            lambda self: getattr(self._instance, attr),  # Getter
            lambda self, value: (
                setattr(self._instance, attr, value)
                if hasattr(type(self._instance), attr) and getattr(type(self._instance), attr).fset
                else None
            ),  # Setter (if allowed)
        )
        setattr(self.__class__, attr, prop)  # Dynamically add the property


# Example usage
class Device:
    def __init__(self):
        self._cvendor = "Dell"
        self.model = "PowerEdge R740"  # Normal attribute, should NOT be exposed as a property

    @property
    def cvendor(self):
        return self._cvendor

    @cvendor.setter
    def cvendor(self, value):
        self._cvendor = value

    @property
    def serial_number(self):
        return "ABC123"  # Read-only property


device = Device()

# Create a PropertyMunch wrapping the Device instance
provides = Munch({"chassis_inventory": PropertyMunch(device)})

# ✅ Only @property attributes are accessible like properties
print(provides["chassis_inventory"].cvendor)  # Output: "Dell"
provides["chassis_inventory"].cvendor = "HP"
print(device.cvendor)  # Output: "HP"

print(provides["chassis_inventory"].serial_number)  # Output: "ABC123"
# provides["chassis_inventory"].serial_number = "XYZ789"  # ❌ Raises AttributeError (no setter)

# ❌ Normal attributes are NOT treated as properties
print(hasattr(provides["chassis_inventory"], "model"))  # Output: False
print("model" in provides["chassis_inventory"])  # Output: False

# ✅ Can still use Munch behavior for custom keys
provides["chassis_inventory"].custom_field = "Extra Data"
print(provides["chassis_inventory"].custom_field)  # Output: "Extra Data"



from typing import TYPE_CHECKING, List
from ekfsm.devices.generic import Device
from ekfsm.devices.utils import retry
from ekfsm.log import ekfsm_logger
from io4edge_client.ssm import Client, Pb

if TYPE_CHECKING:
    from ekfsm.devices.io4edge import IO4Edge

logger = ekfsm_logger(__name__)


class SSM(Device):
    """
    Device class for handling System State Manager (SSM) functionality.

    The SSM manages system state transitions and provides watchdog functionality
    for io4edge devices.
    """

    def __init__(
        self,
        name: str,
        parent: "IO4Edge",
        children: List[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        keepaliveInterval: int = 10000,
        *args,
        **kwargs,
    ):
        logger.debug(
            "Initializing SSM '%s' with parent device %s", name, parent.deviceId
        )

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name
        if service_suffix is not None:
            self.service_suffix = service_suffix
            logger.debug("Using custom service suffix: %s", service_suffix)
        else:
            self.service_suffix = name
            logger.debug("Using default service suffix: %s", name)

        self.deviceId = parent.deviceId
        self.service_addr = f"{self.deviceId}-{self.service_suffix}"
        self.timeout = int(keepaliveInterval / 1000 + 5)

        logger.info(
            "SSM '%s' configured with service address: %s",
            name,
            self.service_addr,
        )

        try:
            self.client = Client(
                self.service_addr, command_timeout=self.timeout, connect=False
            )
            logger.debug("SSM client created for service: %s", self.service_addr)
        except Exception as e:
            logger.error("Failed to create SSM client for %s: %s", self.service_addr, e)
            raise

    @retry()
    def describe(self) -> Pb.ConfigurationDescribeResponse:
        """
        Get the description from the System State Manager functionblock.

        Returns
        -------
        Pb.ConfigurationDescribeResponse
            Description from the SSM functionblock

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        """
        logger.info("Getting SSM description for '%s'", self.name)
        try:
            result = self.client.describe()
            logger.debug("SSM '%s' description retrieved successfully", self.name)
            return result
        except Exception as e:
            logger.error("Failed to get SSM '%s' description: %s", self.name, e)
            raise

    @retry()
    def kick(self) -> None:
        """
        Kick the System State Manager to prevent a timeout.

        This method should be called periodically to indicate that the system
        is still alive and prevent the watchdog from triggering.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        """
        logger.debug("Kicking SSM watchdog for '%s'", self.name)
        try:
            self.client.kick()
            logger.debug("SSM '%s' watchdog kicked successfully", self.name)
        except Exception as e:
            logger.error("Failed to kick SSM '%s' watchdog: %s", self.name, e)
            raise

    @retry()
    def state(self) -> Pb.SystemState:
        """
        Get the current system state from the SSM.

        Returns
        -------
        Pb.SystemState
            Current system state (OFF, ON, ERROR, etc.)

        Raises
        ------
        RuntimeError
            if the command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        """
        logger.info("Getting system state for SSM '%s'", self.name)
        try:
            state = self.client.state
            logger.info("SSM '%s' current state: %s", self.name, state)
            return state
        except Exception as e:
            logger.error("Failed to get SSM '%s' state: %s", self.name, e)
            raise

    @retry()
    def error(self, message: str) -> None:
        """
        Set the system to error state with the given message.

        Parameters
        ----------
        message : str
            Error message to log

        Raises
        ------
        RuntimeError
            if command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        InvalidStateError
            if operation attempted in an invalid state
        UnknownError
            if unknown error occurs while setting error state
        """
        logger.warning("Setting SSM '%s' to error state: %s", self.name, message)
        try:
            self.client.error(message)
            logger.info("SSM '%s' error state set successfully", self.name)
        except Exception as e:
            logger.error("Failed to set SSM '%s' error state: %s", self.name, e)
            raise

    @retry()
    def resolve(self, message: str) -> None:
        """
        Resolve the current error state.

        Parameters
        ----------
        message : str
            Resolution message

        Raises
        ------
        RuntimeError
            if command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        InvalidStateError
            if operation attempted in an invalid state
        UnknownError
            if unknown error occurs while resolving error
        """
        logger.info("Resolving SSM '%s' error state: %s", self.name, message)
        try:
            self.client.resolve(message)
            logger.info("SSM '%s' error resolved successfully", self.name)
        except Exception as e:
            logger.error("Failed to resolve SSM '%s' error: %s", self.name, e)
            raise

    @retry()
    def fatal(self, message: str) -> None:
        """
        Signal a fatal error to the SSM.

        Parameters
        ----------
        message : str
            Fatal error message

        Raises
        ------
        RuntimeError
            if command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        InvalidStateError
            if operation attempted in an invalid state
        UnknownError
            if unknown error occurs signaling fatal error
        """
        logger.critical("Setting SSM '%s' fatal error: %s", self.name, message)
        try:
            self.client.fatal(message)
            logger.info("SSM '%s' fatal error set successfully", self.name)
        except Exception as e:
            logger.error("Failed to set SSM '%s' fatal error: %s", self.name, e)
            raise

    @retry()
    def shutdown(self) -> None:
        """
        Indicate system shutdown to the SSM.

        Raises
        ------
        RuntimeError
            if command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        InvalidStateError
            if operation attempted in an invalid state
        UnknownError
            if unknown error occurs while signaling shutdown
        """
        logger.info("Signaling shutdown to SSM '%s'", self.name)
        try:
            self.client.shutdown()
            logger.info("SSM '%s' shutdown signal sent successfully", self.name)
        except Exception as e:
            logger.error("Failed to signal shutdown to SSM '%s': %s", self.name, e)
            raise

    @retry()
    def on(self) -> None:
        """
        Signal ON state to the SSM.

        Raises
        ------
        RuntimeError
            if command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        InvalidStateError
            if operation attempted in an invalid state
        UnknownError
            if unknown error occurs while turning on
        """
        logger.info("Turning on SSM '%s'", self.name)
        try:
            self.client.on()
            logger.info("SSM '%s' turned on successfully", self.name)
        except Exception as e:
            logger.error("Failed to turn on SSM '%s': %s", self.name, e)
            raise

    @retry()
    def reboot(self) -> None:
        """
        Signal a reboot to the SSM.

        Raises
        ------
        RuntimeError
            if command fails or unhandled response
        TimeoutError
            if the command times out
        ConnectionError
            if not connected to the device
        InvalidStateError
            if operation attempted in an invalid state
        UnknownError
            if unknown error occurs while signaling reboot
        """
        logger.info("Signaling reboot to SSM '%s'", self.name)
        try:
            self.client.reboot()
            logger.info("SSM '%s' reboot signal sent successfully", self.name)
        except Exception as e:
            logger.error("Failed to signal reboot to SSM '%s': %s", self.name, e)
            raise

    def __repr__(self):
        return f"SSM({self.name}; Service Address: {self.service_addr})"



import pprint
from pathlib import Path

from ekfsm.system import System

config = Path(__file__).parent / "srf-fan-test.yaml"
system = System(config)

pprint.pprint(f"System slots {system.slots}")

system.print()

cpu = system["CPU1"]
cpuB = system.cpu1
cpuC = system[0]

assert cpu == cpuB == cpuC

cpu_slot = system.slots["SYSTEM_SLOT"]
cpu_slotB = system.slots.SYSTEM_SLOT
cpu_slotC = system.slots[0]

assert cpu_slot == cpu_slotB == cpu_slotC

cpu.print()
print(f"probing CPU: {cpu.probe()}")
print(f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}")

srf = system[1]
srf.print()
print(f"probing SRF: {srf.probe()}")



from ekfsm.core.components import SysTree
from ekfsm.exceptions import ConfigError
from ekfsm.log import ekfsm_logger

from ..core.sysfs import SysfsDevice
from .generic import Device

logger = ekfsm_logger(__name__)


class MuxChannel(Device):
    """
    A MuxChannel is a device that represents a channel on an I2C multiplexer.
    It is a child of the I2CMux device.
    The MuxChannel device is used to access the I2C bus on the channel.

    Parameters
    ----------
    name
        The name of the device.
    channel_id
        The channel ID of the device.
    parent
        The parent device of the MuxChannel.
    children
        The children of the MuxChannel device. If None, no children are created.
    """

    def __init__(
        self,
        name: str,
        parent: "I2CMux",
        children: list[Device] | None = None,
        abort=False,
        channel_id: int | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(name, parent, children, abort, *args, **kwargs)
        self.channel_id = channel_id

        if parent.sysfs_device is None:
            raise ConfigError(f"{self.name}: Parent I2CMux must have a sysfs_device")
        if not isinstance(self.parent, I2CMux):
            raise ConfigError(f"{self.name}: Parent must be an I2CMux instance")

        path = parent.sysfs_device.path / f"channel-{self.channel_id}"
        self.sysfs_device = SysfsDevice(path, False)


class I2CMux(Device):
    """
    This class represents an I2C multiplexer device.

    Parameters
    ----------
    name
        The name of the device.
    parent
        The parent device of the I2CMux device. If None, no parent is created.
    children
        The children of the I2CMux device. If None, no children are created.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None = None,
        children: list[Device] | None = None,
        abort=False,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr)



import logging
import time
from pathlib import Path

from ekfsm.simctrl import SocketSmbus, enable_simulation, register_gpio_simulations, register_smbus_sim
from ekfsm.system import System

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

enable_simulation(Path(__file__).parent / "sim" / "sys")

# redirecting smbus to socket running in wokwi, where the ESP32 is simulating the smbus via socket
register_smbus_sim(15, 0x28, SocketSmbus("172.28.64.1", 10000))

register_gpio_simulations()

config = Path(__file__).parent / "sim" / "test_system.yaml"
system = System(config)

ccu = system.ccu
ccu.print()

print(f"CCU firmware: {ccu.management.identify_firmware()}")


# def progress_callback(offset):
#     print(f"Loading firmware: {offset}")

# with open("/home/klaus/winDownloads/fw-ccu-00-default.bin", "rb") as f:
#     firmware = f.read()
#     ccu.management.load_firmware(firmware, progress_callback)


params = """
{
    "version": "1.0",
    "parameters": {
        "fan-defrpm":   "4500",
        "shutdn-delay": "40"
    }
}
"""
ccu.management.load_parameterset(params)

print(f"CCU parameters: {ccu.management.get_parameterset()}")

print(f"CCU temperature: {ccu.th.temperature()}C")
print(f"CCU humidity: {ccu.th.humidity()}%")
print(f"CCU VIN voltage: {ccu.vin.voltage()}V")

ccu.fan.push_temperature(-1, 30.0)
for i in range(7):
    for fan in range(2):
        print(f"CCU fan {fan} status: {ccu.fan.fan_status(fan)}")
    time.sleep(1)



import logging
import time
from pathlib import Path

from ekfsm.system import System

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

config = Path(__file__).parent / "zip2-comp1.yaml"
system = System(config)

psu = system.psu

print(f"vendor: {psu.inventory.vendor()}")
print(f"model: {psu.inventory.model()}")
print(f"serial: {psu.inventory.serial()}")
print(f"revision: {psu.inventory.revision()}")


def print_psu_block(block_name, block):
    print(f"block: {block_name}")
    print(f"voltage: {block.voltage()}")
    print(f"current: {block.current()}")
    print(f"status: {block.status()}")


while True:
    print(f"temperature: {psu.th.temperature()}")
    print_psu_block("main (+12V)", psu.main)
    print_psu_block("sby (+5V)", psu.sby)
    time.sleep(0.5)



import pprint
from pathlib import Path

from ekfsm.simctrl import enable_simulation, register_gpio_simulations
from ekfsm.system import System

enable_simulation(Path(__file__).parent / "sim" / "sys")
register_gpio_simulations()


config = Path(__file__).parent / "sim" / "test_system.yaml"
system = System(config)
system.print()
print(system.name)
pprint.pprint(f"System slots {system.slots}")

ccu = system.ccu
assert ccu is not None

eeprom = ccu.mux.ch00.eeprom

eeprom.write_cmodel("SRS-C001")
assert eeprom.cmodel() == "SRS-C001"

eeprom.write_cvendor("EKF Elektronik")
assert eeprom.cvendor() == "EKF Elektronik"

eeprom.write_crevision("2.0")
assert eeprom.crevision() == "2.0"

eeprom.write_unit(1)
assert eeprom.unit() == 1

data = b"\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10"
ccu.custom_eeprom.write(data)
assert ccu.custom_eeprom.read()[0 : len(data)] == data



from pathlib import Path
from typing import Any, List, Tuple, Union

import yaml
from munch import Munch, munchify

from ekfsm.core.components import SysTree
from ekfsm.utils import all_board_cfg_files, find_board_config

from .config import load_config
from .core import HWModule
from .core.slots import Slot, Slots, SlotType
from .exceptions import ConfigError
from .log import ekfsm_logger


class System(SysTree):
    """
    A System represents a CPCI system.

    Once initialised, it will create:
        - a list of boards that are present in the system which can be accessed either by name or by slot number.
        - a list of slots that are present in the system which can be accessed under the slots attribute.

    Visual representation of the system is shown as trees of HW Modules and attached devices.

    Attributes
    ----------
    name
        The name of the system.
    slots
        A dictionary-like object that contains all slots in the system.
    boards
        A list of all boards in the system.
    master
        The master board of the system.
    master_slot_number
        The slot number of the master board.
    config
        The system configuration.


    Accessing boards
    ----------------

    Iterating over the system will iterate over all boards in the system.

    <board_name>
        The board object can be accessed by its name.
    <slot_number>
        The board object can be accessed by its slot number.


    Example
    -------
    >>> from ekfsm.system import System
    >>> system = System("path/to/config.yaml")
    >>> print(system) # Print the system configuration as trees of HWModules
    >>> system.print() # same as above
    >>> cpu = system.cpu # Access the CPU board by its name
    >>> cpu = system[0] # Access the CPU board by its slot index (index as in configuration file)
    >>> print(system.slots) # Print all slots in the system
    >>> print(system.boards) # Print all boards in the system
    >>> for b in system: # Iterate over all boards in the system
    >>>    print(b.name + b.slot.name) # Print the name of the board and the slot it is in
    """

    def __init__(self, config: Path, abort: bool = False) -> None:
        """
        Parameters
        ----------
        config
            Path to the config that specifies the system and how the slots are filled.
        abort
            If True, abort the program if a board cannot be created. If False, leave the slot empty.
            Default is False.
        """
        self.config_path = config
        self.config = load_config(str(self.config_path))
        self.name = self.config.system_config.name

        super().__init__(self.name, abort=abort)

        self.logger = ekfsm_logger(__name__)
        self._init_system(config)
        self._init_slot_attrs()
        self._aggregate_provider_functions()
        self.children = self.boards

    def _init_system(self, config: Path):
        self.slots: Slots = Slots()
        self.boards: List[HWModule] = []

        self.master, self.master_slot_number = self._create_master()
        if self.master is None:
            raise ConfigError("No master board found in system configuration!")

        self.logger.info(f"Master board found in slot {self.master_slot_number}")

        for i, slot_cfg in enumerate(self.config.system_config.slots):
            hwmod: Union[HWModule, Slot, None]
            if i == self.master_slot_number:
                hwmod = self.master
            else:
                hwmod, slot = self.create_hwmodule(slot_cfg, i, self.master)

            if hwmod is not None:
                hwmod.slot.hwmodule = hwmod
                self.boards.append(hwmod)
                self.slots.add(hwmod.slot)
            else:
                self.slots.add(slot)

    def _init_slot_attrs(self):
        for board in self.boards:
            setattr(self, board.instance_name.lower(), board)

    def _aggregate_provider_functions(self):
        if hasattr(self.config.system_config, "aggregates"):
            agg = self.config.system_config.aggregates
            if agg is not None:
                for key, value in agg.items():
                    prv = Munch()
                    for board in self.boards:
                        if hasattr(board, key):
                            prv.update({value: getattr(board, key)})
                    if value in prv.keys():
                        setattr(self, value, prv[value])

    def reload(self):
        """
        Reload the current system configuration.

        Important
        ---------
        This will rebuild all system objects and reinitialize the system tree.
        """
        self.__init__(self.config_path)

    def _create_master(self) -> Tuple[HWModule | None, int]:
        for i, slot in enumerate(self.config.system_config.slots):
            if "attributes" in slot:
                if "is_master" in slot.attributes:
                    if slot.attributes.is_master:
                        master, _ = self.create_hwmodule(slot, i, None)
                        if master is not None:
                            master.master = master  # ???
                            return master, i
                        else:
                            return None, -1
        return None, -1  # ???

    def create_hwmodule(
        self, slot_entry: Munch, slot_number: int, master: HWModule | None
    ) -> Tuple[HWModule | None, Slot]:
        """
        Create HWModule object for the slot.

        Parameters
        ----------
            slot_entry
                The slot entry config (usually part of the system configuration).
            slot_number
                The slot number of the slot.
            master
                The master board of the system.

        Returns
        -------
            HWModule and Slot. HWodule is None if it cannot be created.
        """
        slot = self._create_slot(slot_entry, slot_number, master)
        board_type = slot_entry.desired_hwmodule_type
        board_name = slot_entry.desired_hwmodule_name

        self.logger.debug(
            f"Creating HWModule {board_type} (desired name: {board_name}) in slot {slot.name}"
        )

        if board_type != "":
            # try to create first the desired board
            path = find_board_config(board_type)
            if path is None:
                self.logger.error(
                    f"No board config found for {board_type} (desired name: {board_name})"
                )
                return None, slot

            try:
                hwmod = self._create_hwmodule_from_cfg_file(slot, board_name, path)

            except Exception as e:
                if self.abort:
                    self.logger.error(
                        f"failed to create desired HWModule {board_type} (as {board_name}): {e}. Aborting!"
                    )
                    raise e
                else:
                    self.logger.error(
                        f"failed to create desired HWModule {board_type} (as {board_name}): {e}. Leaving slot empty!"
                    )
                    return None, slot

            # try to probe desired board type
            if hwmod.probe():
                self.logger.info(
                    f"Found desired board type {hwmod.board_type} for slot {slot.name}"
                )
                return hwmod, slot

        # try all other boards types. Maybe someone inserted the wrong board
        self.logger.info(
            f"Probing failed. Trying all other board types for slot {slot.name}"
        )
        for path in all_board_cfg_files():
            try:
                hwmod = self._create_hwmodule_from_cfg_file(slot, board_name, path)
            except ConfigError:
                # slot type not matching, ignore
                # ??? should we log this?
                continue
            except Exception as e:
                self.logger.debug(
                    f"failed to create HWmodule {path} for slot {slot.name}: {e}"
                )
                continue

            if hwmod.probe():
                self.logger.info(
                    f"Found other board type {hwmod.board_type} for slot {slot.name}"
                )
                return hwmod, slot

        return None, slot

    def _create_slot(
        self, slot_entry: Munch, slot_number: int, master: HWModule | None
    ) -> Slot:
        attributes = None
        if "attributes" in slot_entry:
            attributes = slot_entry.attributes

        return Slot(
            slot_entry.name,
            SlotType.from_string(slot_entry.slot_type),
            slot_entry.desired_hwmodule_type,
            slot_entry.desired_hwmodule_name,
            slot_number,
            None,
            master,
            attributes,
        )

    def _create_hwmodule_from_cfg_file(
        self, slot: Slot, board_name: str, path: Path
    ) -> HWModule:
        """
        Try to create a HWModule object from a board config file.
        It does not probe the hardware.

        Returns
        -------
            HWModule object.

        Raises
        ------
            FileNotFoundError
                If the board config file does not exist.
            ConfigError
                If the slot type in the config file does not match the slot type.
            Exception
                If something else went wrong.
        """

        with open(path) as file:
            yaml_data = yaml.safe_load(file)
            cfg = munchify(yaml_data)
            # only instantiate if slot type matches
            if cfg.slot_type != slot.slot_type.to_string():
                raise ConfigError(
                    f"Slot type mismatch for slot {slot.name}: {cfg.slot_type} != {slot.slot_type}"
                )

            hwmod = HWModule(
                instance_name=board_name,
                config=yaml_data,
                slot=slot,
                abort=self.abort,
                parent=self,
            )

            return hwmod

    def get_module_in_slot(self, idx: int) -> HWModule | None:
        """
        Get the HWModule in the given slot.

        Parameters
        ----------
            idx
                The slot index.
        Returns
        -------
            HWModule
                The HWModule in the given slot.
            None
                If no HWModule is present in the given slot.
        """
        return next(
            (
                v.hwmodule
                for k, v in self.slots.items()
                if getattr(v, "number", None) == idx
            ),
            None,
        )

    def get_module_by_name(self, name: str) -> HWModule | None:
        """
        Get the HWModule by its name.

        Parameters
        ----------
            name
                The name of the HWModule.

        Returns
        -------
            HWModule
                The HWModule with the given name.
            None
                If no HWModule is present with the given name.
        """
        return next(
            (
                b
                for b in self.boards
                if getattr(b, "instance_name", None) is not None
                and getattr(b, "instance_name").lower() == name.lower()
            ),
            None,
        )

    def __iter__(self):
        return iter(self.boards)

    def __getitem__(self, key) -> HWModule:
        if isinstance(key, int):
            value = self.get_module_in_slot(key)
        else:
            value = self.get_module_by_name(key)

        if value is None:
            raise KeyError(f"Board {key} not found in system!")

        return value

    def __getattr__(self, name: str) -> Any:
        """Access board by attribute using dot notation"""
        # This fixes mypy error: "... has no object ..."
        if (hw_module := self.get_module_by_name(name)) is not None:
            return hw_module

        raise AttributeError(
            f"'{type(self).__name__}' object has no board with name '{name}'"
        )

    def __repr__(self):
        return f"System (name={self.name})"



from typing import Callable, List
from ekfsm.core.components import HWModule
from ekfsm.devices.generic import Device
from ekfsm.log import ekfsm_logger
import io4edge_client.core.coreclient as CClient
from io4edge_client.binaryiotypeb import Client

from re import sub

logger = ekfsm_logger(__name__)


class IO4Edge(Device):
    """
    Device class for handling IO4Edge devices.

    See https://docs.ci4rail.com/user-docs/io4edge/ for more information.
    """

    def __init__(
        self,
        name: str,
        parent: HWModule | None = None,
        children: List[Device] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        logger.debug("Initializing IO4Edge device '%s'", name)

        super().__init__(name, parent, children, abort, *args, **kwargs)

        attr = self.hw_module.slot.attributes
        if (
            attr is None
            or not hasattr(attr, "slot_coding")
            or getattr(attr, "slot_coding") is None
        ):
            logger.error(
                "Slot attributes for %s are not set or do not contain 'slot_coding'",
                self.hw_module.slot.name,
            )
            raise ValueError(
                f"Slot attributes for {self.hw_module.slot.name} are not set or do not contain 'slot_coding'."
            )
        else:
            geoaddr = int(attr.slot_coding)
            self._geoaddr = geoaddr
            logger.debug("IO4Edge '%s' geo address: %s", name, geoaddr)

        _, module_name = sub(r"-.*$", "", self.hw_module.board_type).split(maxsplit=1)
        self._module_name = module_name
        logger.debug("IO4Edge '%s' module name: %s", name, module_name)

        try:
            self.client = CClient.new_core_client(self.deviceId, connect=False)
            logger.info(
                "IO4Edge '%s' initialized with device ID: %s", name, self.deviceId
            )
        except Exception as e:
            logger.error("Failed to create IO4Edge core client for '%s': %s", name, e)
            raise

    @property
    def deviceId(self) -> str:
        """
        Returns the device ID for the IO4Edge device.
        The device ID is a combination of the module name and the geo address.
        """
        return f"{self._module_name}-geo_addr{self._geoaddr:02d}"

    def identify_firmware(self) -> tuple[str, str]:
        """
        Identify the firmware on the IO4Edge device.

        Returns
        -------
            A tuple containing the firmware title and version.
        """
        response = self.client.identify_firmware()
        return (
            response.title,
            response.version,
        )

    def load_firmware(
        self, cfg: bytes, progress_callback: Callable[[float], None] | None = None
    ) -> None:
        """
        Load firmware onto the IO4Edge device.

        cfg
            Firmware configuration bytes.
        progress_callback
            Optional callback for progress updates.
        """
        self.client.load_firmware(cfg, progress_callback)

    def restart(self) -> None:
        """
        Restart the IO4Edge device.

        Important
        ---------
            This will disconnect the client from the device.
        """
        self.client.restart()

    def load_parameter(self, name: str, value: str) -> None:
        """
        Set a parameter onto the IO4Edge device.

        cfg
            The name of the parameter to load.
        value
            The value to set for the parameter.
        """
        self.client.set_persistent_parameter(name, value)

    def get_parameter(self, name: str) -> str:
        """
        Get a parameter value from the IO4Edge device.

        Returns
            The value of the requested parameter.
        """
        return self.client.get_persistent_parameter(name)

    def __repr__(self):
        return f"{self.name}; DeviceId: {self.deviceId}"


class GPIOArray(Device):
    """
    Device class for handling an io4edge GPIO array.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        keepaliveInterval: int = 10000,
        *args,
        **kwargs,
    ):
        logger.debug(
            "Initializing GPIOArray '%s' with parent device %s", name, parent.deviceId
        )

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name
        if service_suffix is not None:
            self.service_suffix = service_suffix
            logger.debug("Using custom service suffix: %s", service_suffix)
        else:
            self.service_suffix = name
            logger.debug("Using default service suffix: %s", name)

        self.deviceId = parent.deviceId
        self.service_addr = f"{self.deviceId}-{self.service_suffix}"
        self.timeout = int(keepaliveInterval / 1000 + 5)

        logger.info(
            "GPIOArray '%s' configured with service address: %s",
            name,
            self.service_addr,
        )

        try:
            self.client = Client(
                self.service_addr, command_timeout=self.timeout, connect=False
            )
            logger.debug("IO4Edge client created for service: %s", self.service_addr)
        except Exception as e:
            logger.error(
                "Failed to create IO4Edge client for %s: %s", self.service_addr, e
            )
            raise

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import IO4Edge
from ekfsm.devices.utils import retry
from ekfsm.log import ekfsm_logger
from io4edge_client.analogintypeb import Client

logger = ekfsm_logger(__name__)


class ThermalHumidity(Device):
    """
    Device class for handling a thermal humidity sensor.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        *args,
        **kwargs,
    ):
        logger.debug(
            f"Initializing ThermalHumidity sensor '{name}' with parent device {parent.deviceId}"
        )

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        if service_suffix is not None:
            self.service_suffix = service_suffix
            logger.debug(f"Using custom service suffix: {service_suffix}")
        else:
            self.service_suffix = name
            logger.debug(f"Using default service suffix: {name}")

        self.service_addr = f"{parent.deviceId}-{self.service_suffix}"
        logger.info(
            f"ThermalHumidity '{name}' configured with service address: {self.service_addr}"
        )

        try:
            self.client = Client(self.service_addr, connect=False)
            logger.debug(
                f"ThermalHumidity client created for service: {self.service_addr}"
            )
        except Exception as e:
            logger.error(
                f"Failed to create ThermalHumidity client for {self.service_addr}: {e}"
            )
            raise

    @retry()
    def temperature(self) -> float:
        """
        Get the temperature in Celsius.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        logger.info(f"Reading temperature from ThermalHumidity sensor '{self.name}'")
        try:
            temp = self.client.value()[0]
            logger.info(f"ThermalHumidity '{self.name}' temperature: {temp}°C")
            return temp
        except Exception as e:
            logger.error(
                f"Failed to read temperature from ThermalHumidity '{self.name}': {e}"
            )
            raise

    @retry()
    def humidity(self) -> float:
        """
        Get the relative humidity in percent.

        Returns
        -------
            humidity
                relative humidity


        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        logger.info(f"Reading humidity from ThermalHumidity sensor '{self.name}'")
        try:
            temp = self.client.value()[1]
            logger.info(f"ThermalHumidity '{self.name}' humidity: {temp}%")
            return temp
        except Exception as e:
            logger.error(
                f"Failed to read humidity from ThermalHumidity '{self.name}': {e}"
            )
            raise

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



import threading
from typing import Callable, List
from ekfsm.devices.utils import retry
from io4edge_client.binaryiotypeb import Pb
from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import GPIOArray
from ekfsm.log import ekfsm_logger
import io4edge_client.functionblock as fb

logger = ekfsm_logger(__name__)


class Button(Device):
    """
    Device class for handling a single button as part on array.
    """

    def __init__(
        self,
        name: str,
        parent: Device,
        children: List[Device] | None = None,
        abort: bool = False,
        channel_id: int = 0,
        *args,
        **kwargs,
    ):
        logger.debug(f"Initializing Button '{name}' on channel {channel_id}")

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.channel_id = channel_id
        logger.debug(f"Button '{name}' assigned to channel {channel_id}")

        self._handler: Callable | None = None
        logger.info(f"Button '{name}' initialized on channel {channel_id}")

    @property
    def handler(self):
        """
        Handle button events with a callback function.
        """
        return self._handler

    @handler.setter
    def handler(self, func: Callable | None, *args, **kwargs):
        """
        Handle button events with a callback function.

        Parameters
        ----------
        func : Callable | None
            The function to call on button events. If None, no function is called.
        """
        if callable(func):
            self._handler = func
            logger.info(
                f"Handler set for button '{self.name}' on channel {self.channel_id}"
            )
            logger.debug(
                f"Handler function: {func.__name__ if hasattr(func, '__name__') else str(func)}"
            )
        else:
            self._handler = None
            logger.debug(
                f"Handler cleared for button '{self.name}' on channel {self.channel_id}"
            )

    def __repr__(self):
        return f"{self.name}; Channel ID: {self.channel_id}"


class ButtonArray(Device):
    """
    Device class for handling an io4edge gpio based button array.

    To read button events, call the `read` method in a separate thread.

    Note
    ----
        Button handlers are called in the context of the `read` method's thread and need to be set in the Button instances.
    """

    def __init__(
        self,
        name: str,
        parent: GPIOArray,
        children: List[Device] | None = None,
        abort: bool = False,
        keepaliveInterval: int = 10000,
        *args,
        **kwargs,
    ):
        logger.debug(
            f"Initializing ButtonArray '{name}' with parent device {parent.deviceId}"
        )

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        self.service_addr = parent.service_addr
        self.client = parent.client

        logger.info(
            f"ButtonArray '{name}' configured with service address: {self.service_addr}"
        )

        self.subscriptionType = Pb.SubscriptionType.BINARYIOTYPEB_ON_RISING_EDGE
        self.stream_cfg = fb.Pb.StreamControlStart(
            bucketSamples=1,  # 1 sample per bucket, also ein event pro bucket
            keepaliveInterval=keepaliveInterval,
            bufferedSamples=2,  # 2 samples werden gepuffert
            low_latency_mode=True,  # schickt soweit moeglich sofort die Events
        )
        logger.debug(
            "Stream configuration initialized with rising edge subscription and low latency mode"
        )

        # Log button children count
        button_count = sum(1 for child in (children or []) if isinstance(child, Button))
        logger.info(f"ButtonArray '{name}' initialized with {button_count} button(s)")

    def read(self, stop_event: threading.Event | None = None, timeout: float = 1):
        """
        Read all button events and dispatch to handlers.

        Parameters
        ----------
        stop_event : threading.Event, optional
            Event to signal stopping the reading loop. If None, the loop will run indefinitely.
        timeout : float, optional
            Timeout for reading from the stream in seconds. Default is 0.1 seconds.

        Note
        ----
            This method blocks and should be run in a separate thread.
        """
        button_channels = [
            button for button in self.children if isinstance(button, Button)
        ]

        if not button_channels:
            logger.warning(
                f"No button children found in ButtonArray '{self.name}', read operation will have no effect"
            )
            return

        logger.info(
            f"Starting button event reading for {len(button_channels)} buttons on '{self.name}'"
        )
        logger.debug(
            f"Read timeout: {timeout}s, stop_event provided: {stop_event is not None}"
        )
        # Prepare subscription channels
        subscribe_channels = tuple(
            Pb.SubscribeChannel(
                channel=button.channel_id,
                subscriptionType=self.subscriptionType,
            )
            for button in button_channels
        )

        channel_ids = [button.channel_id for button in button_channels]

        try:
            self._button_event_handling(
                stop_event,
                timeout,
                subscribe_channels,
                button_channels,
                channel_ids,
            )
        except Exception as e:
            logger.error(
                f"Failed to establish connection or start stream for ButtonArray '{self.name}': {e}"
            )
            raise

    @retry()
    def _button_event_handling(
        self,
        stop_event: threading.Event | None,
        timeout: float,
        subscribe_channels: tuple,
        button_channels: list,
        channel_ids: list,
    ):
        with self.client as client:
            logger.debug(f"IO4Edge client connected to service: {self.service_addr}")

            logger.debug(
                f"Subscribing to {len(subscribe_channels)} button channels: {channel_ids}"
            )

            client.start_stream(
                Pb.StreamControlStart(subscribeChannel=subscribe_channels),
                self.stream_cfg,
            )
            logger.info(f"Button event stream started for ButtonArray '{self.name}'")

            event_count = 0
            try:
                while not (stop_event and stop_event.is_set()):
                    try:
                        _, samples = client.read_stream(timeout=timeout)

                        for sample in samples.samples:
                            for button in button_channels:
                                pressed = bool(sample.inputs & (1 << button.channel_id))
                                if pressed:
                                    event_count += 1
                                    button_name = getattr(button, "name", "unnamed")
                                    logger.debug(
                                        f"Button press on channel {button.channel_id} ({button_name})"
                                    )

                                    if button.handler:
                                        try:
                                            logger.debug(
                                                f"Calling handler for button on channel {button.channel_id}"
                                            )
                                            button.handler()
                                        except Exception as e:
                                            logger.error(
                                                f"Error in button handler for channel {button.channel_id}: {e}"
                                            )
                                    else:
                                        logger.debug(
                                            f"No handler set for button on channel {button.channel_id}"
                                        )

                    except TimeoutError:
                        # Timeout is expected during normal operation
                        continue
                    except Exception as e:
                        logger.error(f"Error reading button events from stream: {e}")
                        break

            except KeyboardInterrupt:
                logger.info(f"Button reading interrupted for ButtonArray '{self.name}'")
            finally:
                logger.info(
                    f"Button event reading stopped for '{self.name}' after processing {event_count} events"
                )
                if stop_event:
                    stop_event.clear()
                    logger.debug("Stop event cleared")

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



from abc import ABC, abstractmethod
from typing import List


class SimSMBus(ABC):
    @abstractmethod
    def read_word_data(self, cmd: int) -> int:
        pass

    @abstractmethod
    def read_block_data(self, cmd: int) -> List[int]:
        pass

    @abstractmethod
    def write_block_data(self, cmd: int, data: List[int]):
        pass

    @abstractmethod
    def write_byte(self, cmd: int):
        pass

    @abstractmethod
    def write_word_data(self, cmd: int, data: int):
        pass



class ImuSample:
    """
    Class to store IMU data sample

    Parameters
    ----------
    accel
        Accelerometer data in m/s^2, [x, y, z]
    gyro
        Gyroscope data in degrees/s, [x, y, z]
    lost
        True if data was lost before that sample
    """

    def __init__(self, accel: list[float], gyro: list[float], lost: bool):
        self.accel = accel
        self.gyro = gyro
        self.lost = lost



from contextlib import contextmanager


class Connectable:

    def __init__(self, client=None):
        self._client = client
        self._connected = False

    @contextmanager
    def connect(self):
        if not self._connected:
            client = self._client(self.service_addr, command_timeout=self.timeout)
            self._connected = True
        try:
            yield client
        finally:
            client.close()
            del client



#!/usr/bin/env python3
"""
Enhanced profiling analysis script for sq3_profile.prof
"""

import pstats
from pathlib import Path


def analyze_profile():
    profile_file = Path(__file__).parent / "sq3_profile.prof"

    if not profile_file.exists():
        print(f"Profile file not found: {profile_file}")
        print("Run sq3-only.py first to generate the profile")
        return

    stats = pstats.Stats(str(profile_file))
    stats.strip_dirs()

    print("=" * 80)
    print("DETAILED ANALYSIS OF SQ3 PROFILING")
    print("=" * 80)

    print("\n1. EKFSM MODULE ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"ekfsm", 15)

    print("\n2. IO4EDGE_CLIENT MODULE ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"io4edge_client", 15)

    print("\n3. SOCKET/TRANSPORT LAYER ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"socket|transport", 10)

    print("\n4. DISPLAY OPERATIONS ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"display|pixel|image", 10)

    print("\n5. BUTTON OPERATIONS ANALYSIS")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(r"button|read_stream", 10)

    print("\n6. MOST TIME-CONSUMING FUNCTIONS (ALL)")
    print("-" * 40)
    stats.sort_stats("tottime").print_stats(20)

    print("\n7. FUNCTIONS WITH MOST CALLS")
    print("-" * 40)
    stats.sort_stats("ncalls").print_stats(15)

    print("\n8. CUMULATIVE TIME ANALYSIS")
    print("-" * 40)
    stats.sort_stats("cumtime").print_stats(15)


if __name__ == "__main__":
    analyze_profile()



import logging
import time
from pathlib import Path

from ekfsm.system import System

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

config = Path(__file__).parent / "zip2-ccu-only.yaml"
system = System(config)

cpu = system[0]
print(f"probing CPU: {cpu.probe()}")
print(f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}")

ccu = system.ccu


while True:
    sample, more_samples = ccu.imu.sample()
    if sample is not None:
        print(f"accel: {sample.accel}, gyro: {sample.gyro}, lost: {sample.lost} more: {more_samples}")
        time.sleep(0.01)



from enum import Enum


class EkfSmException(Exception):
    """Base class for all exceptions in the EKFSM Library"""

    pass


class ConfigError(EkfSmException):
    """Error in configuration"""

    pass


class SysFSError(EkfSmException):
    """Error while handling sysfs pseudo file system"""

    pass


class GPIOError(EkfSmException):
    """Error while handling GPIO"""

    class ErrorType(Enum):
        INVALID_PIN = "Pin not found"
        NO_MATCHING_DEVICE = "No matching device found"
        NO_MAJOR_MINOR = "No major/minor number found"

    pass

    def __init__(self, error_type: ErrorType, details: str | None = None):
        self.error_type = error_type
        self.details = details
        super().__init__(
            f"{error_type.value}: {details}" if details else error_type.value
        )


class DriverError(EkfSmException):
    """No driver found for device"""

    pass


class HWMonError(EkfSmException):
    """No HwMon entry found for device"""

    pass


class ConversionError(EkfSmException):
    """Failed to convert"""

    pass


class UnsupportedModeError(EkfSmException):
    """Format not supported"""

    pass


class FirmwareNodeError(EkfSmException):
    """Error while handlig firmware node"""

    pass


class DataCorruptionError(EkfSmException):
    """Error while handling data corruption"""

    def __init__(self, details: str | None = None):
        self.details = details
        super().__init__(
            f"Data corruption: {details}" if details else "Data corruption"
        )


class AcquisitionError(EkfSmException):
    """Error while handling data acquisition"""

    pass



"""
Comprehensive CCTV system test script.
Tests SSM functionality, system components, LED control, GPIO operations, and button handling.
"""

import pprint
import threading
import logging
from pathlib import Path
from time import sleep
from ekfsm.system import System

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

# Initialize system
# NOTE: Ensure that "cctv-comp1.yaml" is in the same directory as this script and properly configured for your system.
config = Path(__file__).parent / "cctv-comp1.yaml"
system = System(config, abort=True)

print("=" * 60)
print("=== CCTV Comprehensive Test Script ===")
print("=" * 60)

# ============================================================================
# SECTION 1: System and Component Tests
# ============================================================================
print("\n### SECTION 1: System and Component Tests ###\n")

print("System slots:")
pprint.pprint(system.slots)

system.print()

# Test different ways to access CPU
cpu = system["CPU"]
cpuB = system.cpu
cpuC = system[0]

assert cpu == cpuB == cpuC
print("\nCPU access methods verified (all point to same object)")

cpu.print()
print(f"\nProbing CPU: {cpu.probe()}")
print(
    f"Inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
)

# Get SMC and i4edge interface
smc = system.smc
smc.print()

i4e = smc.i4e
print(f"\ni4edge interface: {i4e}")

# ============================================================================
# SECTION 2: SSM (System State Manager) Tests
# ============================================================================
print("\n### SECTION 2: SSM (System State Manager) Tests ###\n")

ssm = i4e.ssm
print(f"SSM Device: {ssm}")
print(f"Service Address: {ssm.service_addr}\n")

# Test 1: Get SSM description
print("Test 1: Getting SSM description...")
try:
    description = ssm.describe()
    print(f"  Description: {description}")
except Exception as e:
    print(f"  Error: {e}")

# Test 2: Get current system state
print("\nTest 2: Getting current system state...")
try:
    state = ssm.state()
    print(f"  Current state: {state}")
except Exception as e:
    print(f"  Error: {e}")

# Test 3: Turn system ON
print("\nTest 3: Turning system ON...")
try:
    ssm.on()
    print("  System turned ON successfully")
    sleep(1)
    state = ssm.state()
    print(f"  New state: {state}")
except Exception as e:
    print(f"  Error: {e}")

# Test 4: SSM watchdog kicks
print("\nTest 4: Kicking SSM watchdog 5 times...")
for i in range(5):
    try:
        ssm.kick()
        print(f"  SSM Kick {i+1}/5 successful")
        sleep(0.5)
    except Exception as e:
        print(f"  SSM Kick {i+1}/5 failed: {e}")
        break

# Test 5: Set and resolve error state
print("\nTest 5: Testing error state...")
try:
    ssm.error("Test error message")
    print("  Error state set successfully")
    sleep(1)
    state = ssm.state()
    print(f"  State after error: {state}")
    ssm.resolve("Test error resolved")
    print("  Error resolved successfully")
    sleep(1)
    state = ssm.state()
    print(f"  State after resolve: {state}")
except Exception as e:
    print(f"  Error: {e}")

# ============================================================================
# SECTION 3: LED Control Tests
# ============================================================================
print("\n### SECTION 3: LED Control Tests ###\n")

print("Setting and verifying LED states...")
try:
    # LED 2 test
    i4e.leds.led2.set(0, True)
    led2 = i4e.leds.led2.get()
    assert led2 == (0, True)
    print(f"  LED2: brightness=0, on=True -> {led2} ✓")

    # LED 5 test
    i4e.leds.led5.set(3, True)
    led5 = i4e.leds.led5.get()
    assert led5 == (3, True)
    print(f"  LED5: brightness=3, on=True -> {led5} ✓")

    # LED 3 test
    i4e.leds.led3.set(5, False)
    led3 = i4e.leds.led3.get()
    assert led3 == (5, False)
    print(f"  LED3: brightness=5, on=False -> {led3} ✓")
    print("\nAll LED tests passed!")
except Exception as e:
    print(f"  LED test error: {e}")

# ============================================================================
# SECTION 4: GPIO and Button Tests
# ============================================================================
print("\n### SECTION 4: GPIO and Button Tests ###\n")

button_array = i4e.gpios.buttons
pw_toggle = i4e.gpios.power
redriver_toggle = i4e.gpios.redriver

eject = button_array.eject

# Set up button handler
eject.handler = lambda: print("  [BUTTON EVENT] Eject button pressed!")

# Start button monitoring thread
print("Starting button monitoring thread...")
stop_event = threading.Event()
button_thread = threading.Thread(target=button_array.read, args=(stop_event,))
button_thread.start()
print("  Button thread started (press eject button to test)\n")

# ============================================================================
# SECTION 5: Comprehensive Loop Test
# ============================================================================
print("### SECTION 5: Comprehensive Loop Test ###\n")
print("Running 30 iterations of LED and GPIO tests...")
for i in range(30):
    try:
        # LED cycling tests
        i4e.leds.led1.set(1, True)
        led1 = i4e.leds.led1.get()
        assert led1 == (1, True)
        i4e.leds.led4.set(2, True)
        led4 = i4e.leds.led4.get()
        assert led4 == (2, True)
        i4e.leds.led6.set(4, False)
        led6 = i4e.leds.led6.get()
        assert led6 == (4, False)

        # GPIO toggle tests
        pw_toggle.off()
        pw_toggle.on()
        redriver_toggle.off()
        redriver_toggle.on()

        if (i + 1) % 10 == 0:
            print(f"  Iteration {i+1}/30 complete")

    except Exception as e:
        print(f"  Error in iteration {i+1}: {e}")
        break

print("\nLoop test complete!")

# Stop button monitoring thread
print("\nStopping button monitoring thread...")
stop_event.set()
button_thread.join()
print("  Button thread stopped")

# ============================================================================
# Test Summary
# ============================================================================
print("\n" + "=" * 60)
print("=== All Tests Complete ===")
print("=" * 60)
print("\nNote: Shutdown and fatal tests are commented out to avoid system disruption.")
print("Uncomment them if you need to test those functions.\n")

# Uncomment to test shutdown:
# print("\nTest: Testing shutdown...")
# ssm.shutdown()

# Uncomment to test fatal error:
# print("\nTest: Testing fatal error...")
ssm.fatal("Test fatal error")



from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import IO4Edge
from ekfsm.devices.utils import retry
from ekfsm.log import ekfsm_logger
from io4edge_client.pixelDisplay import Client

logger = ekfsm_logger(__name__)


class PixelDisplay(Device):
    """
    Device class for handling a pixel display.
    """

    def __init__(
        self,
        name: str,
        parent: IO4Edge,
        children: list[Device] | None = None,
        abort: bool = False,
        service_suffix: str | None = None,
        *args,
        **kwargs,
    ):
        logger.debug(
            f"Initializing PixelDisplay '{name}' with parent device {parent.deviceId}"
        )

        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.name = name

        if service_suffix is not None:
            self.service_suffix = service_suffix
            logger.debug(f"Using custom service suffix: {service_suffix}")
        else:
            self.service_suffix = name
            logger.debug(f"Using default service suffix: {name}")

        self.service_addr = f"{parent.deviceId}-{self.service_suffix}"
        logger.info(
            f"PixelDisplay '{name}' configured with service address: {self.service_addr}"
        )

        try:
            self.client = Client(self.service_addr, connect=False)
            logger.debug(
                f"PixelDisplay client created for service: {self.service_addr}"
            )
        except Exception as e:
            logger.error(
                f"Failed to create PixelDisplay client for {self.service_addr}: {e}"
            )
            raise

    @retry()
    def describe(self) -> dict:
        """
        Returns a description of the pixel display.
        """
        logger.debug(f"Getting PixelDisplay description for '{self.name}'")
        try:
            describe = self.client.describe()
            desc = {
                "height": describe.height_pixel,
                "width": describe.width_pixel,
                "max_num_of_pixel": describe.max_num_of_pixel,
            }
            logger.debug(f"PixelDisplay '{self.name}' description: {desc}")
            return desc
        except Exception as e:
            logger.error(
                f"Failed to get PixelDisplay description for '{self.name}': {e}"
            )
            raise

    @property
    def height(self) -> int:
        """
        Returns the height of the pixel display in pixels.
        """
        return self.describe()["height"]

    @property
    def width(self) -> int:
        """
        Returns the width of the pixel display in pixels.
        """
        return self.describe()["width"]

    @retry()
    def off(self) -> None:
        """
        Turn off the pixel display.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        logger.info(f"Turning off PixelDisplay '{self.name}'")
        try:
            self.client.set_display_off()
            logger.debug(f"PixelDisplay '{self.name}' successfully turned off")
        except Exception as e:
            logger.error(f"Failed to turn off PixelDisplay '{self.name}': {e}")
            raise

    @retry()
    def display_image(self, pixel: list[list[tuple[int, int, int]]]) -> None:
        """
        Display an image on the pixel display.

        Parameters
        ----------
        pixel : list[list[tuple[int, int, int]]]
            Pixel data to be displayed on the pixel display.

        Raises
        ------
        RuntimeError
            if the command fails
        TimeoutError
            if the command times out
        """
        logger.info(f"Displaying image on PixelDisplay '{self.name}'")
        try:
            with self.client as client:
                logger.debug(f"Sending pixel data to PixelDisplay '{self.name}'")
                for i in range(0, 320, 16):
                    pix_area = []
                    for k in range(0, 16):
                        for j in range(0, 240):
                            pix_area.append(pixel[j][i + k])
                    client.set_pixel_area(0, i, 239, pix_area)
                logger.debug(
                    f"Image successfully displayed on PixelDisplay '{self.name}'"
                )
        except Exception as e:
            logger.error(f"Failed to display image on PixelDisplay '{self.name}': {e}")
            raise

    def __repr__(self):
        return f"{self.name}; Service Address: {self.service_addr}"



from ekfsm.core.components import SysTree
from ekfsm.devices.generic import Device

from .gpio import GPIOExpander


class EKFSurLed(GPIOExpander):
    """
    A class to represent the EKF-SUR-LED devices.
    """

    def __init__(
        self,
        name: str,
        parent: SysTree | None,
        children: list["Device"] | None = None,
        abort: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, None, abort, *args, **kwargs)

    def __str__(self) -> str:
        return (
            f"EKFSurLed - GPIO Number: {self.number}; "
            f"sysfs_path: {self.sysfs_device.path if self.sysfs_device else ''}"
        )

    def set(self, led: int, color: str):
        """
        Set the color of a LED.

        Parameters
        ----------
        led : int
            The LED number (0 or 1).
        color : str
            The color of the LED.
            Possible values: "off", "red", "blue", "green", "yellow", "purple", "cyan", "white"
        """
        # 3-color LEDs,
        # 0: Red
        # 1: Blue
        # 2: Green

        if color == "off":
            state = [False, False, False]
        elif color == "red":
            state = [True, False, False]
        elif color == "blue":
            state = [False, True, False]
        elif color == "green":
            state = [False, False, True]
        elif color == "yellow":
            state = [True, True, False]
        elif color == "purple":
            state = [True, False, True]
        elif color == "cyan":
            state = [False, True, True]
        elif color == "white":
            state = [True, True, True]
        else:
            raise ValueError(f"Invalid color: {color}")

        if led < 0 or led > 1:
            raise ValueError(f"Invalid led number: {led}")

        for i in range(3):
            self.set_direction(i + 4 * led, True)
            # Active low
            self.set_pin(i + 4 * led, False if state[i] else True)



import logging
import pprint
from pathlib import Path
import cProfile
import pstats

from ekfsm.system import System

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

# Initialize system first (without profiling the setup)
config = Path(__file__).parent / "sq3-only.yaml"
system = System(config, abort=True)

pprint.pprint(f"System slots {system.slots}")

system.print()

cpu = system["CPU"]
cpuB = system.cpu
cpuC = system[0]

assert cpu == cpuB == cpuC

print(cpu.hwmon.cputemp())

eeprom = cpu.eeprom
eeprom.manufactured_at()

cpu_slot = system.slots["SYSTEM_SLOT"]
cpu_slotB = system.slots.SYSTEM_SLOT
cpu_slotC = system.slots[0]

assert cpu_slot == cpu_slotB == cpu_slotC

cpu.print()
print(f"probing CPU: {cpu.probe()}")
print(
    f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
)

sq3 = system["info"]
i4e = sq3.bmc

# Start profiling HIER - nach System-Setup, vor io4edge_client Aufrufen
profiler = cProfile.Profile(builtins=False)
profiler.enable()

fw_title, fw_name = i4e.identify_firmware()

print(f"Firmware: {fw_title} {fw_name}")

pixel = i4e.display

pixel.off()
pixel.display_image(str(Path(__file__).parent / "sim/SQ3.png"))


profiler.disable()

print("\n" + "=" * 80)
print("PROFILING RESULTS - ekfsm and io4edge_client modules")
print("=" * 80)

# Create stats object and filter for our modules
stats = pstats.Stats(profiler)
stats.strip_dirs()  # clean paths

# Print comprehensive profiling results
print("\n1. TOP FUNCTIONS BY TOTAL TIME (including subcalls)")
stats.sort_stats("tottime").print_stats(
    r"ekfsm|io4edge_client|client\.py|display|button", 30
)

print("\n2. TOP FUNCTIONS BY CUMULATIVE TIME (time + subcalls)")
stats.sort_stats("cumtime").print_stats(
    r"ekfsm|io4edge_client|client\.py|display|button", 30
)

print("\n3. TOP FUNCTIONS BY NUMBER OF CALLS")
stats.sort_stats("ncalls").print_stats(
    r"ekfsm|io4edge_client|client\.py|display|button", 20
)

print("\n3a. ALL FUNCTIONS - DEBUGGING VIEW (to see what's actually being called)")
stats.sort_stats("tottime").print_stats(50)  # Show top 50 functions overall

print("\n4. CALLERS/CALLEES ANALYSIS FOR TOP 5 FUNCTIONS")
# Get the top 5 functions by total time
top_funcs = stats.get_stats_profile().func_profiles
filtered_funcs = {
    k: v
    for k, v in top_funcs.items()
    if any(mod in str(k) for mod in ["ekfsm", "io4edge_client"])
}
sorted_funcs = sorted(filtered_funcs.items(), key=lambda x: x[1].tottime, reverse=True)

for i, (func_key, func_stats) in enumerate(sorted_funcs[:5]):
    print(f"\n--- Function #{i+1}: {func_key} ---")
    print(f"Total time: {func_stats.tottime:.6f}s, Calls: {func_stats.ncalls}")
    stats.print_callers(func_key)

# Save detailed profile to file for later analysis
profile_file = Path(__file__).parent / "sq3_profile.prof"
profiler.dump_stats(str(profile_file))
print(f"\n5. Detailed profile saved to: {profile_file}")
print("   Use: python -m pstats sq3_profile.prof")
print("   Or: snakeviz sq3_profile.prof (install: pip install snakeviz)")

print("\n" + "=" * 80)



from functools import wraps
from time import sleep
from crcmod.predefined import Crc
from ekfsm.log import ekfsm_logger

logger = ekfsm_logger(__name__)


def get_crc16_xmodem(data: bytes) -> int:
    crc16_xmodem = Crc("xmodem")
    crc16_xmodem.update(data)
    return crc16_xmodem.crcValue


def retry(max_attempts=5, delay=0.5):
    """
    Retry decorator.

    Decorator that retries a function a number of times before giving up.

    This is useful for functions that may fail due to transient errors.

    Note
    ----
    This is needed for certain PMBus commands that may fail due to transient errors
    because page switching timing is not effectively handled by older kernel versions.

    Important
    ---------
    This decorator is _not_ thread-safe across multiple ekfsm processes. Unfortunately,
    we cannot use fcntl or flock syscalls with files on virtual filesystems like sysfs.

    Parameters
    ----------
    max_attempts
        The maximum number of attempts before giving up.
    delay
        The delay in seconds between attempts.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_attempts:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempts += 1
                    if attempts == max_attempts:
                        logger.exception(
                            f"Failed to execute {func.__name__} after {max_attempts} attempts: {e}"
                        )
                        raise e
                    logger.info(f"Retrying execution of {func.__name__} in {delay}s...")
                sleep(delay)

        return wrapper

    return decorator



from ekfsm.devices.utils import retry
from io4edge_client.binaryiotypeb import Client
from ekfsm.devices.generic import Device
from ekfsm.devices.io4edge import GPIOArray
from ekfsm.log import ekfsm_logger

logger = ekfsm_logger(__name__)


class BinaryToggle(Device):
    """
    Device class for handling a binary toggle switch.
    """

    def __init__(
        self,
        name: str,
        parent: GPIOArray,
        children: list[Device] | None = None,
        abort: bool = False,
        channel_id: int = 0,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, children, abort, *args, **kwargs)

        self.channel_id = channel_id
        self.name = name
        self.service_addr = parent.service_addr
        self.client: Client = parent.client

        logger.debug(
            f"Initializing BinaryToggle '{name}' with parent device {parent.deviceId}"
        )
        logger.info(
            f"BinaryToggle '{name}' configured with service address: {self.service_addr} on channel {channel_id}"
        )

    @retry()
    def set(self, state: bool):
        """
        Set the state of the toggle switch.

        Parameters
        ----------
            state
                state to set. a "true" state turns on the toggle switch, a "false" state turns it off.
        """
        logger.info(
            f"Setting BinaryToggle '{self.name}' on channel {self.channel_id} to state {state}"
        )
        self.client.set_output(self.channel_id, state)
        logger.info(
            f"BinaryToggle '{self.name}' on channel {self.channel_id} set to state {state}"
        )

    @retry()
    def get(self) -> bool:
        """
        Get the current state of the toggle switch.

        Returns
            The current state of the toggle switch.
        """
        state = self.client.get_input(self.channel_id)
        logger.info(
            f"Retrieved state {state} for BinaryToggle '{self.name}' on channel {self.channel_id}"
        )
        return state

    def on(self):
        """
        Turn the toggle switch on.
        """
        self.set(True)

    def off(self):
        """
        Turn the toggle switch off.
        """
        self.set(False)

    def __repr__(self):
        return f"{self.name}; Channel ID: {self.channel_id}"
